/* dump.c - A tool to get data from a Mazda head unit data bus and convert
            to traditional serial data that can be read by any computer.

    This code was written for an atmel ATmega168.
    
    The data is obtained using a pin-changed interrupt on pin 14, it's stored
    in a buffer until an inactivity timer overflows, and is then transmitted
    using the USART on pin 3 (19200 baud, 1 stop bit, no parity data).
    
    The data is sent one command at a time followed by a footer. The footer is
    two bytes: 0xFF, then the number of bits contained in the command.

    Build: avr-gcc -DF_CPU=8000000 -std=gnu99 -mmcu=atmega168 -Os dump.c -o dump
    Create hex file: avr-objcopy -O ihex dump dump.hex
    Program: avrdude -c dasa -p m168 -P /dev/ttyS0 -U flash:w:dump.hex
*/

#include <avr/io.h>
#include <avr/interrupt.h>

#define buffer_size 16
volatile uint8_t buffer[buffer_size];
volatile uint8_t buffer_pos = 0;
volatile uint8_t byte_pos = 0;


#define UBRR_CALC(cpu_freq, baud) (cpu_freq/16/baud - 1)

void init_usart( uint16_t ubrr_value )
{
    UBRR0H = (ubrr_value >> 8);
    UBRR0L = (ubrr_value & 0xff);
    
    UCSR0B = _BV(RXEN0) | _BV(TXEN0); // Enable Rx and Tx
}

void send_byte( uint8_t byte )
{
    loop_until_bit_is_set( UCSR0A, UDRE0 ); // Wait for buffer to be empty
    UDR0 = byte;                            // Load data into the buffer
}

int main()
{
    // Set up serial I/O
    // 9600 baud is too slow to transmit some of the larger commands in the
    // ~10ms window we're given. At 19200 baud we don't run into any problems.
    init_usart( UBRR_CALC(F_CPU, 19200) );
    
    // 8 bit timer (prescaler = 256)
    TCCR0A = _BV(WGM01);
    TCCR0B = _BV(CS02);
    OCR0A = 145; // Trigger an interrupt every ~4.6ms
    TIMSK0 = _BV(OCIE0A);    
    
    // Interrupt on pin change
    PCICR = _BV(PCIE0);
    PCMSK0 = _BV(PCINT0); // We want pin 14 (PB0)
    
    // Enable interrupts
    sei();
    
    // Do nothing, everything is handled by interrupts
    while(1);
    return 0;
}

// Pin change interrupt
ISR(PCINT0_vect)
{
    uint8_t elapsed_time = TCNT0;  // Store the timer value
    uint8_t pin = PINB & _BV(PB0); // Store the pin value
    
    // If the bus is pulled to ground, a bit is starting to be transmitted
    // so the timer is reset and we can return while we wait for the 
    // bus to be pulled back to 5V.
    if (!pin){
        TCNT0 = 0;
        return;
    }
    
    // A logical 1 is when the bus is held low for ~1.7ms so we accept
    // anything from ~1.4ms to ~2ms as a logical 1.
    if (elapsed_time > 43 && elapsed_time < 63)
        buffer[buffer_pos] |= 1<<(7-byte_pos);
    
    // Anything else is assumed to be a 0, increase the byte position
    byte_pos++;
    
    // We've filled up a byte, move to the next one
    if (byte_pos > 7){
        byte_pos = 0;
        buffer_pos++;
        buffer[buffer_pos] = 0; // very important, clear the old value
    }
}

// Timer compare match interrupt
ISR(TIMER0_COMPA_vect)
{
    if (!buffer_pos){
        // clean up any bits that have been stored because of noise
        // (very useful on startup)
        byte_pos = 0;
        buffer[0] = 0;
        return;
    }
    
    if (byte_pos) // sometimes a byte is only half-filled, make sure to send it
        buffer_pos++;
    
    for (uint8_t i=0; i<buffer_pos; i++)
        send_byte(buffer[i]);
    
    // Send the footer
    send_byte(0xff);
    send_byte(buffer_pos*8 + byte_pos);
    
    buffer_pos = 0;
    byte_pos = 0;
    buffer[0] = 0;
}


